﻿using System;
using System.Threading;
using System.Threading.Tasks;
using DevExpress.ExpressApp;
using DevExpress.ExpressApp.Utils;
using Xafari;
using Xafari.MQ;
using Xafari.Server;
using XafariMQSample.Module;

namespace XafariMQSample
{
    class MessageHandler:IMessageHandler
    {
        private static readonly Random TimeGenerator=new Random();
        private string _key;
        private readonly object _messageKey;
        private CancellationToken _cancellationToken;
        private CancellationTokenSource _source;

        public event EventHandler OnCompleted = delegate { };
        public event EventHandler OnStartedThread = delegate { };
        public event EventHandler OnStoppedThread = delegate { };
        public event EventHandler Completed;
        public event EventHandler ThreadStarted;
        public event EventHandler ThreadStopped;

        public string Key
        {
            get
            {
                if (string.IsNullOrEmpty(_key))
                {
                    using (var os = MQManager.Instance.CreateObjectSpace())
                    {
                        var message = os.GetObjectByKey<MQMessage>(_messageKey);
                        _key = os.GetKeyValueAsString(message);
                    }
                }
                return _key;
            }
        }

        public object MessageKey => throw new NotImplementedException();

        public MessageHandler(XafApplication application, object messageKey)
        {
            Guard.ArgumentNotNull(messageKey, "messageKey");

            _messageKey = messageKey;
            _source = new CancellationTokenSource();
        }

        public void StartHandler()
        {
            _cancellationToken = _source.Token;

            LockReportMessage();

            var handlerTask = Task.Factory.StartNew(ProcessMessage, _cancellationToken, TaskCreationOptions.LongRunning, TaskScheduler.Current);
            Task.Factory.ContinueWhenAll(new[] { handlerTask }, tasks => FinishProcessing(), _cancellationToken);
        }

        public void CancelHandler()
        {
            _source.Cancel();
            OnProcessedMessage();
        }

        private void LockReportMessage()
        {
            if (!_cancellationToken.IsCancellationRequested)
            {
                using (var objectSpace = MQManager.Instance.CreateObjectSpace())
                {
                    var updatedMessage = objectSpace.GetObjectByKey<MQMessage>(_messageKey);
                    if (updatedMessage.CurrentStatus != MessageStatus.Processing)
                    {
                        updatedMessage.CurrentStatus = MessageStatus.Processing;
                        objectSpace.CommitChanges();
                    }
                }
            }
        }

        private void ProcessMessage()
        {
            OnStartedThread(Thread.CurrentThread, EventArgs.Empty);

            if (!_cancellationToken.IsCancellationRequested)
            {
                int threadSleepTime = 1000*TimeGenerator.Next(10);
                Thread.Sleep(threadSleepTime);

                using (var objectSpace = MQManager.Instance.CreateObjectSpace())
                {
                    var message = objectSpace.GetObjectByKey<MQMessage>(_messageKey);
                    var dbObjectSpace = XafariModule.Application.ObjectSpaceProvider.CreateObjectSpace();
                    var messageData = (MessageData) message.MessageData.GetTargetObject(objectSpace);
                    var log = dbObjectSpace.CreateObject<MessageLog>();
                    log.Index = messageData.Index;
                    log.CreationTime = DateTime.Now;
                    log.Name = message.Description;
                    log.ThreadSleepTime = threadSleepTime;
                    dbObjectSpace.CommitChanges();

                    message.CurrentStatus = MessageStatus.Finished;
                }
            }
        }

        private void FinishProcessing()
        {
            OnCompleted(this, EventArgs.Empty);
            OnStoppedThread(Thread.CurrentThread, EventArgs.Empty);
            Dispose();
        }

        private void OnProcessedMessage()
        {
            _source = new CancellationTokenSource();
            FinishProcessing();
        }

        #region IDisposable implementation

        private bool _isDisposed = false;
        public void Dispose()
        {
            if (_isDisposed)
                return;
            _isDisposed = true;
            _key = String.Empty;
            _source.Cancel();
        }

        #endregion
    }
}
